(* This theory defines raw, typed HOAS, split into atomic and normal 
   forms, along with well-formedness relative to typed contexts.
*)

theory HOAS
imports Nominal Name
begin

text {* The type of higher-order abstract syntax terms, 
  parameterized by the type of constants *}

nominal_datatype hoas = 
    VAR "var"
  | APP "hoas" "hoas"
  | LAM "\<guillemotleft>var\<guillemotright>hoas"
  | CST "string"



text {* Types, parameterized over the base types *}

nominal_datatype  hoas_ty = BASE "string"
  | FUN "hoas_ty" "hoas_ty" ("_ ~> _" [60,60] 60)

types hoas_sig = "(string * hoas_ty) list"
types hoas_ctx = "(var * hoas_ty) list"



lemma hoas_ty_cases: "(\<exists>\<tau>1 \<tau>2. b=\<tau>1 ~> \<tau>2) \<or> (\<exists> c. b=BASE  c)"
  apply(nominal_induct b rule:hoas_ty.strong_inducts)
  by(auto)

instantiation hoas_ty :: size
begin

nominal_primrec size_hoas_ty
where
  "size (BASE c) = 1"
| "size (T\<^isub>1~>T\<^isub>2) = size T\<^isub>1 + size T\<^isub>2"
by (rule TrueI)+

instance ..

end
lemma hoas_ty_size_greater_zero[simp]:
  fixes T::"hoas_ty"
  shows "size T > 0"
by (nominal_induct rule: hoas_ty.strong_induct) (simp_all)



lemma fresh_ty[simp]: 
  fixes x::"var"
  and ty::"hoas_ty"
  shows   "x \<sharp> ty"
  apply(induct ty rule: hoas_ty.induct)
  by(simp_all add: fresh_string)
  

lemma fresh_sig[simp]: 
  fixes x::"var"
  and sig::"(string * hoas_ty) list"
  shows   "x \<sharp> sig"
  apply(induct sig)
  apply (simp_all add:fresh_list_cons fresh_list_nil fresh_string )
  apply(induct_tac a)
  apply (simp_all add:fresh_list_cons fresh_list_nil fresh_string )
  done



inductive valid_ctx  :: "hoas_ctx \<Rightarrow> bool"
where 
  valid_ctx_emp[simp] : "valid_ctx []"
| valid_ctx_cons[intro!] : "\<lbrakk>x \<sharp> X; valid_ctx X\<rbrakk> \<Longrightarrow> valid_ctx ((x,\<tau>)#X)"

lemma fresh_not_in_ctx: 
  fixes x::var
  and \<Gamma>::hoas_ctx
  shows "x \<sharp> \<Gamma> \<Longrightarrow> ((x,\<tau>) \<notin> set \<Gamma>)"
  apply(induct \<Gamma>)
  by(auto simp add:fresh_prod fresh_list_cons fresh_atm)


lemma ctx_det: "valid_ctx \<Gamma> ==> (x,\<tau>) \<in> set \<Gamma> ==> (x,\<tau>') \<in> set \<Gamma> ==> \<tau> = \<tau>'"
  apply(induct \<Gamma> rule:valid_ctx.induct) 
  apply(auto dest: fresh_not_in_ctx)
  done

lemma valid_sub_valid:
  "valid_ctx (\<Gamma>' @[(x,\<tau>)]@ \<Gamma>) \<Longrightarrow> valid_ctx \<Gamma>"
  apply(induct \<Gamma>')
  apply(auto dest: valid_ctx.cases)
  done

lemma valid_sub:
  "valid_ctx (\<Gamma>'@[(x,\<tau>)]@\<Gamma>) \<Longrightarrow> valid_ctx (\<Gamma>'@\<Gamma>)"
  apply(induct \<Gamma>')
  apply(auto elim: valid_ctx.cases simp add: fresh_list_append fresh_list_cons)
  done

lemma valid_not_in:
  "valid_ctx (\<Gamma>' @ (x, \<tau>) # \<Gamma>) \<Longrightarrow> x \<sharp> \<Gamma>' \<and> x \<sharp> \<Gamma>"
  apply(induct \<Gamma>')
  apply(erule valid_ctx.cases, simp_all add:fresh_list_nil)
  apply(erule valid_ctx.cases, simp_all)
  apply(auto simp add:fresh_list_cons fresh_list_append fresh_prod fresh_atm fresh_ty)
  done




fun
  extends :: "hoas_ctx \<Rightarrow> hoas_ctx \<Rightarrow> bool" ("_ extends _" [60,60] 60) 
where
  "\<Gamma>' extends \<Gamma> = (set \<Gamma> \<subseteq> set \<Gamma>' \<and> valid_ctx \<Gamma>')" 


fun 
  list_dom:: "('a * 'b) list \<Rightarrow> 'a set" 
where
  "list_dom [] = {}"
| "list_dom((x,\<tau>)#L) = {x} \<union> list_dom L"

inductive valid_sig :: "hoas_sig => bool" 
where
  valid_sig_nil : "valid_sig []"
| valid_sig_cons: "\<lbrakk>valid_sig \<Sigma>; c \<notin> list_dom \<Sigma>\<rbrakk> \<Longrightarrow> valid_sig ((c,\<tau>)#\<Sigma>)"

lemma fresh_not_in_sig: 
  fixes x::string
  and \<Sigma>::hoas_sig
  shows "x \<notin> list_dom \<Sigma> \<Longrightarrow> ((x,\<tau>) \<notin> set \<Sigma>)"
  by(induct \<Sigma>)(auto)


lemma sig_det: "valid_sig \<Sigma> ==> (c,\<tau>) \<in> set \<Sigma> ==> (c,\<tau>') \<in> set \<Sigma> ==> \<tau> = \<tau>'"
  by(induct \<Sigma> rule:valid_sig.induct) (auto dest:fresh_not_in_sig)





nominal_primrec 
subst :: "hoas \<Rightarrow> var \<Rightarrow> hoas \<Rightarrow> hoas" ("_{_::=_}" [100,100,100] 100)
where
  "(VAR x){y::=t} = (if x =y then t else (VAR x))"
| "(APP t1 t2){y::=t} = APP (t1{y::=t}) (t2{y::=t})"
| "x\<sharp>(y,t) \<Longrightarrow> (LAM x t0){y::=t} = LAM x (t0{y::=t})"
| "(CST c){y::=t} = CST c"
  apply(finite_guess+)
  apply(rule TrueI)+
  apply(simp add:abs_fresh)
  apply(fresh_guess)+
  done

lemma subst_eqvt[eqvt]:
  fixes pi:: "var prm"
  shows "pi\<bullet>(t1{b::=t2}) = (pi\<bullet>t1){(pi\<bullet>b)::=(pi\<bullet>t2)}"
  apply(nominal_induct t1 avoiding: b t2 rule: hoas.strong_induct)
  apply(auto simp add: perm_bij fresh_prod fresh_atm fresh_bij)
  done

lemma subst_supp: 
  shows "supp(t1{a::=t2}) \<subseteq> (((supp(t1)-{a})\<union>supp(t2))::var set)"
apply(nominal_induct t1 avoiding: a t2 rule: hoas.strong_induct)
apply(auto simp add: hoas.supp supp_atm fresh_prod abs_supp supp_string)
apply(blast)+
done



lemma subst_fresh : "x \<sharp> n \<Longrightarrow> x \<sharp> m{x::=n}"
  apply(nominal_induct m avoiding: x n rule:hoas.strong_induct)
  apply(simp add:hoas.fresh fresh_atm abs_fresh fresh_string)+
  done

lemma subst_fresh2: 
  fixes x::var
  shows "\<lbrakk>x \<sharp> M; x \<sharp> N \<rbrakk> \<Longrightarrow> x \<sharp> M{y::=N}"
  apply(nominal_induct M avoiding: x y N rule:hoas.strong_induct)
  apply(auto simp add:abs_fresh) 
  done

lemma subst_fresh_id: "x \<sharp> m \<Longrightarrow> m{x::=n} = m"
  apply(nominal_induct m avoiding: x n rule: hoas.strong_induct)
  by(auto simp add:abs_fresh fresh_atm)


lemma subst_comm: 
  "\<lbrakk>x \<sharp> y; x \<sharp> N'\<rbrakk> \<Longrightarrow> M{x::=N}{y::=N'} = M{y::=N'}{x::=N{y::=N'}}"
  apply(nominal_induct M avoiding: x y N N' rule: hoas.strong_induct)
  apply(auto simp add:fresh_atm subst_fresh_id subst_fresh2)
  done


lemma subst_id: "M{x::=VAR x} = M"
  apply(nominal_induct M avoiding: x rule: hoas.strong_induct)
  by(auto)


lemma subst_swap: "\<lbrakk>y \<sharp> M;y \<sharp> N\<rbrakk> \<Longrightarrow> ([(y,x)]\<bullet> M){y::=N} = M{x::=N}"
  apply(nominal_induct M avoiding: x y N rule: hoas.strong_induct)
  by(auto simp add: hoas.inject calc_atm fresh_atm abs_fresh perm_string)

lemma subst_swap2:
  shows "y \<sharp> m \<Longrightarrow> m{x::=VAR y} =  [(x, y)] \<bullet> m"
  apply(nominal_induct m avoiding: x y rule: hoas.strong_induct)
  apply(auto simp add:hoas.inject calc_atm fresh_atm abs_fresh perm_string)
  done



types hoas_subst = "(var * hoas) list"

fun 
  lookup:: "var => hoas_subst => hoas"
  where
    "lookup x [] = VAR x"
  | "lookup x ((y,M)#\<theta>) = (if x = y then M else lookup x \<theta>)"


lemma lookup_eqvt[eqvt]:
  fixes pi::"var prm"
  shows "pi\<bullet>(lookup x \<theta>) = lookup (pi\<bullet>x) (pi\<bullet>\<theta>)"
by (induct \<theta>) (auto simp add: perm_bij)





nominal_primrec 
ssubst :: "hoas \<Rightarrow> hoas_subst \<Rightarrow> hoas" ("_{_}" [100,100] 900)
where
  "(VAR x){\<theta>} = (lookup x \<theta>)"
| "(APP t1 t2){\<theta>} = APP (t1{\<theta>}) (t2{\<theta>})"
| "x\<sharp>(\<theta>) \<Longrightarrow> (LAM x t0){\<theta>} = LAM x (t0{\<theta>})"
| "(CST c){\<theta>} = CST c"
apply(finite_guess)+
apply(rule TrueI)+
apply(simp add: abs_fresh)+
apply(fresh_guess)+  
done


lemma lookup_fresh_merge:
  shows "x \<sharp> \<theta> \<Longrightarrow> (lookup y \<theta>) {x::=N} = lookup y ((x,N)#\<theta>)"
  apply(induct \<theta>)
  apply(simp_all add:fresh_prod fresh_list_cons)
  apply(clarsimp)
  apply(auto simp add:fresh_prod fresh_atm subst_fresh_id)
  done

lemma subst_fresh_merge:
  fixes M::hoas
  shows "x \<sharp> \<theta> \<Longrightarrow> M {\<theta>} {x::=N} = M {((x,N)#\<theta>)}"
  apply(nominal_induct M avoiding: \<theta> N x rule:hoas.strong_inducts)
  apply(auto simp add:lookup_fresh_merge)
  apply(subgoal_tac "var \<sharp> ((x,N)#\<theta>)",simp_all add:fresh_prod fresh_list_cons)
  done


lemma lookup_fresh_id:
  "x \<sharp> \<theta> ==> lookup x \<theta> = VAR x"
  apply(induct \<theta>)
  apply(simp)
  apply(clarsimp)
  apply(simp add:fresh_list_cons fresh_prod fresh_atm)
  done

lemma lookup_fresh:
  fixes x::"var"
  shows "x \<sharp> (v,\<theta>) ==> x \<sharp> lookup v \<theta>"
  apply(induct \<theta>)
  apply(auto simp add:fresh_prod fresh_list_cons)
  done

lemma ssubst_fresh:
  fixes x::var
  shows "x \<sharp> (M,\<theta>) ==> x \<sharp> (ssubst M \<theta>)"
  apply(nominal_induct M avoiding: x \<theta> rule:hoas.strong_inducts)
  apply(auto simp add:fresh_prod lookup_fresh abs_fresh)
  done


lemma subst_fresh_ssubst:
  "x \<sharp> (\<theta>, u) \<Longrightarrow> t{x::=u}{\<theta>} = t{\<theta>}{x::=u{\<theta>}}"
  apply(nominal_induct t avoiding: x \<theta> u rule:hoas.strong_inducts)
  
  apply(case_tac "x=var")
  apply(simp_all add:fresh_prod lookup_fresh lookup_fresh_merge lookup_fresh_id ssubst_fresh)
  done


fun 
  id_sub:: "hoas_ctx \<Rightarrow> hoas_subst" 
where
  "id_sub [] = []"
| "id_sub((x,A)#\<Gamma>) = ((x,VAR x)#(id_sub \<Gamma>))"


lemma id_lookup_id:
  "lookup x (id_sub \<Gamma>) = VAR x"
  by(induct \<Gamma>) (auto)


lemma id_subst_fresh: "(x::var) \<sharp> \<Gamma> \<Longrightarrow> x \<sharp> id_sub \<Gamma>"
  by(induct \<Gamma>) (auto simp add:fresh_list_cons fresh_list_nil)


lemma id_subst_id: 
  "M{id_sub \<Gamma>} = M"
  by(nominal_induct M avoiding: \<Gamma> rule:hoas.strong_inducts)
  (auto simp add:id_lookup_id id_subst_fresh)






nominal_primrec (invariant: "\<lambda>s::var set. finite s") 
  fvs :: "hoas \<Rightarrow> var set"
where
  "fvs (VAR x) = {x}"
| "fvs (APP m n) = fvs m \<union> fvs n"
| "fvs (LAM x m) = fvs m - {x}"
| "fvs (CST c) = {}"
apply(finite_guess)+
apply(simp)+ 
apply(simp add: fresh_def)
apply(simp add: supp_of_fin_sets[OF pt_var_inst, OF at_var_inst, OF fs_at_inst[OF at_var_inst]])
apply(simp add: supp_atm)
apply(blast)
apply(fresh_guess)+
done

lemma finite_fvs : "finite (fvs x)"
  apply(induct x rule:hoas.induct)
  apply(simp_all)
  done

lemma finite_fvs_list :"\<exists> X. fvs x = set X"
  apply(subgoal_tac "finite (fvs x)")
  prefer 2 apply(rule finite_fvs)
  apply(erule finite_induct)
  apply(blast)
  apply(auto)
  apply(rule_tac x="x#X" in exI, auto)
  done



nominal_primrec height :: "hoas \<Rightarrow> nat" ("|_|")
where
  "height (VAR x) = 0"
| "height (APP t1 t2) = (max (height t1) (height t2)) + 1"
| "height (CST c) = 0"
| "height (LAM x t) = height(t)+1"
apply(finite_guess)+
apply(simp add:fresh_nat)+ 
apply(fresh_guess)+
done

theorem hoas_height_induct: "(\<forall> x. (\<forall> y. height y < height x \<longrightarrow> P y) \<longrightarrow> P x) \<Longrightarrow> P t"
  apply(subgoal_tac "\<forall> n. (\<forall>y. height(y) < n \<longrightarrow> P y)")
  apply(erule_tac x="height(t)+1" in allE)
  apply(erule_tac x="t" in allE)
  back
  apply(erule mp)
  apply(arith)

  apply(rule allI)
  apply(induct_tac n rule:nat_less_induct)
  apply(clarify)
  apply(erule_tac x="height y" in allE)
  apply(erule impE, assumption)
  apply(erule_tac x="y" in allE)
  apply(blast)
  done

text {* Beta/eta equivalence *}

inductive equiv :: "hoas \<Rightarrow> hoas \<Rightarrow> bool" ("_ \<approx> _" [60,60] 60)
where
  equiv_var[intro!]: "VAR (x::var) \<approx> VAR (x::var)"
| equiv_cst[intro!]: "CST c \<approx> CST c"
| equiv_app[intro!]: "\<lbrakk>t \<approx> t'; u \<approx> u'\<rbrakk> \<Longrightarrow> APP t u \<approx> APP t' u'" 
| equiv_lam[intro!]: "t \<approx> t' \<Longrightarrow> LAM x t \<approx> LAM x t'"
| equiv_beta[intro]: "x\<sharp> u \<Longrightarrow> APP (LAM x t) u \<approx> t{x::=u}"
| equiv_eta[intro]: "x\<sharp>t \<Longrightarrow> t \<approx> LAM x (APP t (VAR x))"
| equiv_refl[intro]: "t \<approx> t"
| equiv_sym: "t \<approx> u \<Longrightarrow> u \<approx> t"
| equiv_trans: "\<lbrakk>t \<approx> u; u \<approx> v\<rbrakk> \<Longrightarrow> t \<approx> v" 

equivariance equiv

nominal_inductive equiv
  apply(auto simp add:  abs_fresh  subst_fresh)
  done


text {* Primitive well-formedness.  As with Raw, we only consider 
  primitive HOAS terms well-formed if no shadowing occurs *}

inductive wf :: " hoas_sig \<Rightarrow> 
                  hoas_ctx \<Rightarrow> 
                  hoas \<Rightarrow> hoas_ty \<Rightarrow> bool"
  ("_,_ \<turnstile> _ : _" [60,60,60,60] 60)
where
  wf_VAR[intro!] : "(x,ty) \<in> set ctx \<Longrightarrow> wf sig ctx (VAR x) ty"
| wf_APP[intro!] : "\<lbrakk>wf sig ctx t1 (FUN ty1 ty2);
                    wf sig ctx t2 ty1\<rbrakk> \<Longrightarrow> 
                    wf sig ctx (APP t1 t2) ty2"
| wf_CST[intro!] : "(c,ty) \<in> set sig \<Longrightarrow> wf sig ctx (CST c) ty"
| wf_LAM[intro!] : "\<lbrakk>x \<sharp> ctx;wf sig ((x,ty1)#ctx) t1 ty2\<rbrakk> \<Longrightarrow> 
                    wf sig ctx (LAM x t1) (ty1 ~> ty2)"

equivariance wf

nominal_inductive wf
  apply(auto simp add: abs_fresh )
  done

inductive_cases wf_inv:
  "wf sig ctx (VAR x) ty"
  "wf sig ctx (APP t1 t2) ty2"
  "wf sig ctx (CST c) ty"
  "wf sig ctx (LAM x t1) (ty1 ~> ty2)"



lemma wf_fresh:
  fixes x::var
  shows "\<Sigma>,\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> x \<sharp> \<Gamma> \<Longrightarrow> x \<sharp> M"
  apply(nominal_induct rule: wf.strong_inducts)
  apply(auto simp add:fresh_prod fresh_list_cons abs_fresh fresh_string fresh_atm fresh_not_in_ctx)
  done


lemma weakening:
  shows "\<Sigma>,\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> \<Gamma>' extends \<Gamma> ==> \<Sigma>,\<Gamma>' \<turnstile> M : \<tau>"
  apply(nominal_induct avoiding: \<Gamma>' rule:wf.strong_inducts)
  apply(auto)
  apply(drule_tac x="(x, ty1) # \<Gamma>'" in meta_spec,auto)
  done


lemma substitution_general:
  shows "\<Sigma>,\<Gamma>'@(x,\<tau>')#\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> valid_ctx (\<Gamma>'@(x,\<tau>')#\<Gamma>)  \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> N : \<tau>' \<Longrightarrow>  \<Sigma>,\<Gamma>'@\<Gamma> \<turnstile> M{x::=N} : \<tau>"
  apply(nominal_induct \<Sigma> \<Gamma>\<equiv>"\<Gamma>'@(x,\<tau>')#\<Gamma>" M \<tau> avoiding: \<Gamma> \<Gamma>' x N  rule:wf.strong_inducts)
  apply(clarsimp)
  apply(frule valid_not_in)
  apply(auto simp add:fresh_list_append)
  apply(erule weakening, auto intro:valid_sub)
  apply(auto simp add:fresh_not_in_ctx)
  apply(drule_tac x="\<Gamma>" in meta_spec)
  apply(drule_tac x="(x,ty1) # \<Gamma>'" in meta_spec)
  apply(drule_tac x="xa" in meta_spec)
  apply(drule_tac x="N" in meta_spec)
  apply(drule meta_mp,auto simp add:fresh_list_append)
  done
  

lemma substitution:
  shows "\<Sigma>,(x,\<tau>')#\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow>x \<sharp> \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> N : \<tau>' \<Longrightarrow>  \<Sigma>,\<Gamma> \<turnstile> M{x::=N} : \<tau>"
  apply(subgoal_tac "\<Sigma>,[]@\<Gamma> \<turnstile> M{x::=N} : \<tau>",simp)
  apply(rule substitution_general)
  apply(auto)
  done



(* Standard defn of weak head reduction *)

inductive whr :: "hoas \<Rightarrow> hoas \<Rightarrow> bool" ("_ \<leadsto> _" [60,60] 60)
where
  whr_beta[intro!] : "x\<sharp> N \<Longrightarrow> APP (LAM x M) N \<leadsto> M{x::=N}"
| whr_app[intro!] : "M \<leadsto> M' \<Longrightarrow> APP M N \<leadsto> APP M' N" 


equivariance whr

nominal_inductive whr
  by(auto simp add:fresh_string abs_fresh subst_fresh)+


lemma whr_beta_strong: "APP (LAM x M) N \<leadsto> M{x::=N}"
proof -
  obtain y::"var" where fc: "y\<sharp>(x,M,N)" by (rule_tac exists_fresh) (auto simp add: fs_var1)
  have "APP (LAM x M) N = APP (LAM y ([(y,x)]\<bullet> M)) N" using fc 
    by (auto simp add: hoas.inject alpha' fresh_atm fresh_prod)
  also have "\<dots> \<leadsto>  ([(y,x)]\<bullet> M){y::=N}" using fc by (rule_tac whr_beta) (simp)
  also have "\<dots> = M{x::=N}" using fc by (simp add: subst_swap)
  finally show "APP (LAM x M) N \<leadsto> M{x::=N}" by simp
qed

lemma whr_id: "APP (LAM x M) (VAR x) \<leadsto> M"
  apply(subgoal_tac "APP (LAM x M) (VAR x) \<leadsto> M{x::=VAR x}")
  apply(simp add:subst_id)
  apply(rule whr_beta_strong)
  done


lemma whr_equiv: 
  "M \<leadsto> M' \<Longrightarrow> M \<approx> M'"
  apply(nominal_induct M M' rule:whr.strong_inducts)
  apply(auto)
  done


lemma whr_det: "M \<leadsto> N \<Longrightarrow> M \<leadsto> N' \<Longrightarrow> N = N'"
  apply(nominal_induct M N avoiding: N' rule: whr.strong_inducts)
  (* case 1 *)
  apply(erule_tac x="x" in  whr.strong_cases)
  apply(simp add:abs_fresh hoas.inject alpha)
  apply(simp add:abs_fresh hoas.inject alpha,clarify)
  apply(erule whr.cases, simp,simp)
  (* case 2 *)
  apply(erule_tac ?a1.0="APP M N" in whr.cases) 
  apply(simp (no_asm_use) only:hoas.inject)
  apply(clarify)
  apply(erule whr.cases, simp_all (no_asm_use))
  apply(simp (no_asm_use) only:hoas.inject)
  apply(clarify,blast)
  done

lemma whr_subst: 
  "M \<leadsto> M' \<Longrightarrow> M{x::=N} \<leadsto> M'{x::=N}"
  apply(nominal_induct M M' avoiding: x N rule:whr.strong_inducts)
  apply(auto )
  apply(subgoal_tac "M{x::=N}{xa::=Na} = M{xa::=Na}{x::=N{xa::=Na}}")
  apply(auto intro:subst_fresh2 subst_comm)
  done

lemma whr_fresh : "M \<leadsto> N \<Longrightarrow> (x::var) \<sharp> M \<Longrightarrow> x \<sharp> N"
  apply(nominal_induct M N avoiding: x rule:whr.strong_inducts)
  apply(auto simp add:abs_fresh subst_fresh subst_fresh2)
  done


(* Inductive definition of weak head normal forms and "neutral" forms *)
(* A neutral term is an app starting with a constant or var.
   A WHNF is either neutral or a lam,bda or an app whose head is neutral. *)

(* The main reason for introducing whnfs is to make it easy to tell when
    a term cannot take a whr step.  This rules out many vacuous cases
    in one step by appealing to theorem whnf_whr_impos. *)

inductive whnf :: "hoas \<Rightarrow> bool" 
  and neut :: "hoas \<Rightarrow> bool" 
where
  neut_var[intro] : "neut (VAR x)"
| neut_cst[intro] : "neut (CST c)"
| neut_app[intro] : "neut m \<Longrightarrow> neut (APP m n)"
| whnf_neut[intro] : "neut m \<Longrightarrow> whnf m"
| whnf_lam[intro] : "whnf (LAM x m)"


equivariance whnf

nominal_inductive whnf
  by(auto simp add:fresh_string abs_fresh subst_fresh)+


lemma whr_whnf_impos':
  "whnf M \<Longrightarrow> M \<leadsto> M' \<Longrightarrow> False"
  "neut M \<Longrightarrow> M \<leadsto> M' \<Longrightarrow> False"
  apply(nominal_induct M and M arbitrary: M' and M' rule: whnf_neut.strong_inducts)
  apply(auto elim: whr.cases)
  apply(erule whr.cases)
  apply(simp add:hoas.inject,clarify)
  apply(erule neut.cases,simp_all)
  apply(simp add:hoas.inject,clarify,blast)
  done

theorem whr_whnf_impos: 
  "M \<leadsto> M' \<Longrightarrow> whnf M \<Longrightarrow> False"
  apply(auto intro: whr_whnf_impos')
  done

(*
theorem not_whr_whnf:
  "\<not> (\<exists> N. M \<leadsto> N) \<Longrightarrow> whnf M"
  apply(nominal_induct M  rule: hoas.strong_inducts)
  apply(auto)
  apply(
  sorry

theorem whnf_correct:
  "whnf(M) = (\<not> (\<exists> M'. M \<leadsto> M'))"
  apply(auto intro: whr_whnf_impos not_whr_whnf)
  done

*)


text {* Beta/eta equivalence, typed *}

inductive wf_equiv :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas \<Rightarrow> hoas \<Rightarrow> hoas_ty \<Rightarrow> bool" 
  ("_,_ \<turnstile> _ \<approx> _ : _" [60,60,60,60,60] 60)
where
  wf_equiv_var[intro!]: "(x,\<tau>) \<in> set \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> VAR x \<approx> VAR x : \<tau>"
| wf_equiv_cst[intro!]: "(c,\<tau>) \<in> set \<Sigma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> CST c \<approx> CST c : \<tau>"
| wf_equiv_app[intro!]: "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> t \<approx> t' : \<tau> ~> \<tau>'; \<Sigma>,\<Gamma> \<turnstile> u \<approx> u' : \<tau>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile>APP t u \<approx> APP t' u' : \<tau>'" 
| wf_equiv_lam[intro!]: "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma> \<turnstile> t \<approx> t' : \<tau>';x\<sharp>\<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> LAM x t \<approx> LAM x t' : \<tau> ~> \<tau>'"
| wf_equiv_beta[intro]: "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> APP (LAM x t) u : \<tau>;x\<sharp>(\<Gamma>,u)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> APP (LAM x t) u \<approx> t{x::=u} : \<tau>"
| wf_equiv_eta[intro]: "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> t : \<tau> ~> \<tau>';x\<sharp>(\<Gamma>,t)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> t \<approx> LAM x (APP t (VAR x)) : \<tau> ~> \<tau>'"
| wf_equiv_sym: "\<Sigma>,\<Gamma> \<turnstile> t \<approx> u : \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> u \<approx> t : \<tau>"
| wf_equiv_trans: "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> t \<approx> u : \<tau>; \<Sigma>,\<Gamma> \<turnstile> u \<approx> v : \<tau>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> t \<approx> v : \<tau>" 

equivariance wf_equiv

nominal_inductive wf_equiv
  apply(auto simp add:  abs_fresh subst_fresh)
  done





lemma wf_equiv_equiv: 
  "\<Sigma>,\<Gamma> \<turnstile> M \<approx> N : \<tau> \<Longrightarrow> M \<approx> N"
  apply(nominal_induct rule:wf_equiv.strong_inducts)
  apply(auto)
  apply(blast intro: equiv_sym)
  apply(blast intro: equiv_trans)
  done

lemma wf_equiv_refl:
  "\<Sigma>,\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<approx> M : \<tau>"
  apply(nominal_induct rule:wf.strong_inducts)
  by(auto)


lemma wf_equiv_wf_:
  "\<Sigma>,\<Gamma> \<turnstile> M \<approx> N : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M : \<tau> \<and> \<Sigma>,\<Gamma> \<turnstile> N : \<tau>"
  apply(nominal_induct rule:wf_equiv.strong_inducts)
  apply(auto)
  apply(erule wf.cases,simp_all add:hoas.inject) 
  apply(erule_tac x="x" in wf.strong_cases,simp_all add:hoas.inject abs_fresh ) 
  apply(clarsimp,simp add:abs_fresh hoas.inject hoas_ty.inject alpha,clarsimp)
  apply(erule substitution,simp_all)
  apply(rule weakening,auto)
  done

lemma wf_equiv_wf:
  "\<Sigma>,\<Gamma> \<turnstile> M \<approx> N : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M : \<tau> "
  "\<Sigma>,\<Gamma> \<turnstile> M \<approx> N : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> N : \<tau> "
  by(auto dest:wf_equiv_wf_)



lemma wf_equiv_var_inv:
  "\<Sigma>,\<Gamma> \<turnstile> VAR x \<approx> VAR x : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> (x,\<tau>) \<in> set \<Gamma>"
  apply(drule wf_equiv_wf)
  apply(simp)
  apply(erule wf_inv, simp add:hoas.inject)
  done

lemma wf_equiv_cst_inv:
 "\<Sigma>,\<Gamma> \<turnstile> CST c \<approx> CST c : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> (c,\<tau>) \<in> set \<Sigma>"
  apply(drule wf_equiv_wf)
  apply(simp)
  apply(erule wf_inv, simp add:hoas.inject)
  done



lemma wf_equiv_weakening:
  "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> M \<approx> N : \<tau>; \<Gamma>' extends \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>' \<turnstile> M \<approx> N : \<tau>"
  apply(nominal_induct \<Sigma> \<Gamma> M N \<tau> avoiding: \<Gamma>' rule:wf_equiv.strong_inducts)
  apply(auto intro:wf_equiv.intros) 
  apply(drule_tac x="(x, \<tau>) # \<Gamma>'" in meta_spec,auto) 
  apply(rule wf_equiv_beta, rule weakening, auto)
  apply(rule wf_equiv_eta, rule weakening, auto)
  done


lemma wf_equiv_subst:
  "\<lbrakk>\<Sigma>,\<Gamma>'@[(x,\<tau>)]@\<Gamma> \<turnstile> M \<approx> N : \<tau>'; \<Sigma>,\<Gamma> \<turnstile> t : \<tau>; valid_ctx (\<Gamma>'@[(x,\<tau>)]@\<Gamma>) \<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>'@\<Gamma> \<turnstile> M{x::=t} \<approx> N{x::=t} : \<tau>'"
  apply(nominal_induct \<Sigma> \<Gamma>\<equiv>"\<Gamma>'@[(x,\<tau>)]@\<Gamma>" M N \<tau>' avoiding: \<Gamma> x \<Gamma>' t \<tau> rule:wf_equiv.strong_inducts)
  (* case 1 *)
  apply(case_tac "x=xa")
  apply(subgoal_tac "\<tau> = \<tau>'")
  apply(clarsimp)
  apply(rule wf_equiv_refl)
  apply(erule weakening)
  apply(simp add:valid_sub,clarify)
  apply(clarify)
  apply(erule ctx_det)
  apply(assumption)
  apply(simp)
    (* case 1a *)
  apply(clarsimp)
    (* case 2 *)
  apply(clarsimp)
    (* case 3 *)
  apply(clarsimp)
  apply(blast intro:wf_equiv.intros)
    (* case 4 *)
  apply(clarsimp)
  apply(rule wf_equiv.intros)
  apply(drule_tac x="\<Gamma>'" in meta_spec)
  apply(drule_tac x="xa" in meta_spec)
  apply(drule_tac x="(x, \<tau>) # \<Gamma>'a" in meta_spec)
  apply(drule_tac x="ta" in meta_spec)
  apply(drule_tac x="\<tau>''" in meta_spec)
  apply(simp (no_asm_use))
  apply(clarsimp)
  apply(subgoal_tac "valid_ctx ((x, \<tau>) # \<Gamma>'a @ (xa, \<tau>'') # \<Gamma>')")
  prefer 2 apply(blast intro:valid_ctx.intros)
  apply(simp,simp add:fresh_list_append)
    (* case 5 *)
  apply(clarsimp)
  apply(subgoal_tac "t{x::=u}{xa::=ta} = t{xa::=ta}{x::=u{xa::=ta}}",clarsimp)
  prefer 2 apply(rule subst_comm,simp,simp)
  apply(rule wf_equiv.intros)
  apply(subgoal_tac "\<Sigma>,\<Gamma>'a @ \<Gamma>' \<turnstile> (APP (LAM x t) u){xa::=ta} : \<tau>",simp add:subst_fresh)
  apply(erule substitution_general)
  apply(assumption)+
  apply(simp add:fresh_prod fresh_list_append subst_fresh2)
    (* case 6 *)
  apply(clarify)
  apply(simp add:fresh_atm)
  apply(rule wf_equiv.intros)
  apply(erule substitution_general)
  apply(assumption)+
  apply(simp add:fresh_prod fresh_list_append subst_fresh2)
    (* case 7 *)
  apply(blast intro:wf_equiv_sym)
(* case 8 *)
  apply(blast intro:wf_equiv_trans)
  done

lemma wf_equiv_subst1:
  "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma> \<turnstile> M \<approx> N : \<tau>; \<Sigma>,\<Gamma> \<turnstile> t : \<tau>; valid_ctx \<Gamma>; x \<sharp> \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M{x::=t} \<approx> N{x::=t} : \<tau> "
  apply(subgoal_tac "\<Sigma>,[]@\<Gamma> \<turnstile> M{x::=t} \<approx> N{x::=t} : \<tau>",simp)
  apply(rule wf_equiv_subst,auto)
  done


lemma wf_equiv_wf_subst:
  "\<lbrakk>\<Sigma>,\<Gamma>'@[(x,\<tau>')]@\<Gamma> \<turnstile> M : \<tau>; \<Sigma>,\<Gamma> \<turnstile> t \<approx> u : \<tau>'; valid_ctx (\<Gamma>'@[(x,\<tau>')]@\<Gamma>)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>'@\<Gamma> \<turnstile> M{x::=t} \<approx> M{x::=u} : \<tau> "
  apply(nominal_induct \<Sigma> \<Gamma>\<equiv>"\<Gamma>'@[(x,\<tau>')]@\<Gamma>" M \<tau> avoiding: \<Gamma> x \<Gamma>' t u \<tau>' rule:wf.strong_inducts)
  apply(case_tac "x=xa")
  apply(clarify)
  apply(subgoal_tac "ty=\<tau>'")
  prefer 2 apply(erule ctx_det,assumption,clarsimp)
  apply(clarsimp, erule wf_equiv_weakening, auto intro:valid_sub)
  apply(blast)
  apply(blast)
  apply(drule_tac x="\<Gamma>" in meta_spec)
  apply(drule_tac x="xa" in meta_spec)
  apply(drule_tac x="(x,ty1) # \<Gamma>'" in meta_spec)
  apply(drule_tac x="t" in meta_spec)
  apply(drule_tac x="u" in meta_spec)
  apply(drule_tac x="\<tau>'" in meta_spec)
  apply(simp (no_asm_use))
  apply(drule meta_mp,assumption)
  apply(drule meta_mp,auto simp add:fresh_list_append)
  done


lemma wf_equiv_equiv_subst:
  "\<lbrakk>\<Sigma>,\<Gamma>'@[(x,\<tau>')]@\<Gamma> \<turnstile> M \<approx> N : \<tau>; \<Sigma>,\<Gamma> \<turnstile> t \<approx> u : \<tau>'; valid_ctx (\<Gamma>'@[(x,\<tau>')]@\<Gamma>)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma>'@\<Gamma> \<turnstile> M{x::=t} \<approx> N{x::=u} : \<tau> "
  apply(nominal_induct \<Sigma> \<Gamma>\<equiv>"\<Gamma>'@[(x,\<tau>')]@\<Gamma>" M N \<tau> avoiding: \<Gamma> x \<Gamma>' t u \<tau>' rule:wf_equiv.strong_inducts)
  (* case 1 *)
  apply(case_tac "x=xa")
  apply(subgoal_tac "\<tau> = \<tau>'")
  apply(clarsimp)
  apply(erule wf_equiv_weakening)
  apply(simp add:valid_sub,clarify)
  apply(clarify)
  apply(erule_tac x="xa" in ctx_det)
  apply(simp,simp)
    (* case 1b *)
  apply(clarsimp)
    (* case 2 *)
  apply(clarsimp)
    (* case 3 *)
  apply(clarsimp)
  apply(blast intro:wf_equiv.intros)
    (* case 4 *)
  apply(clarsimp)
  apply(rule wf_equiv.intros)
  apply(drule_tac x="\<Gamma>'" in meta_spec)
  apply(drule_tac x="xa" in meta_spec)
  apply(drule_tac x="(x, \<tau>) # \<Gamma>'a" in meta_spec)
  apply(drule_tac x="ta" in meta_spec)
  apply(drule_tac x="u" in meta_spec)  
  apply(drule_tac x="\<tau>'a" in meta_spec)
  apply(simp (no_asm_use))
  apply(clarsimp)
  apply(subgoal_tac "valid_ctx ((x, \<tau>) # \<Gamma>'a @ (xa, \<tau>'a) # \<Gamma>')")
  prefer 2 apply(blast intro:valid_ctx.intros)
  apply(simp,simp add:fresh_list_append)
    (* case 5 *)
  apply(clarsimp)

  apply(rule wf_equiv_trans, rule wf_equiv_beta)
  apply(subgoal_tac "\<Sigma>,\<Gamma>'a @ \<Gamma>' \<turnstile> (APP (LAM x t) u){xa::=ta} : \<tau>",simp add:subst_fresh)
  apply(erule substitution_general)
  apply(assumption)
  apply(erule wf_equiv_wf)
  apply(erule valid_sub_valid[simplified])
  apply(simp add:fresh_prod fresh_list_append subst_fresh2)
  apply(subgoal_tac "t{xa::=ta}{x::=u{xa::=ta}} = t{x::=u}{xa::=ta}")
  prefer 2 apply(simp add: subst_comm)
  apply(clarsimp)
  apply(erule wf_inv, simp add:hoas.inject,clarify)
  apply(erule_tac x="x" in wf.strong_cases, simp,simp,simp)
  apply(simp add:abs_fresh alpha  hoas.inject hoas_ty.inject, clarify)
  apply(subgoal_tac "sig,\<Gamma>'a @ (xa, \<tau>') # \<Gamma>' \<turnstile> t1a{x::=t2} : ty2")
  prefer 2 apply(erule substitution,simp, simp add:fresh_prod,assumption)  
  apply(erule wf_equiv_wf_subst[simplified],simp,simp)
    (* case 6 *)
  apply(clarify)
  apply(simp add:fresh_atm)
  apply(subgoal_tac "\<Sigma>,\<Gamma>'a @ \<Gamma>' \<turnstile> t{xa::=ta} \<approx> t{xa::=u} : \<tau> ~> \<tau>'")
  prefer 2 apply(erule wf_equiv_wf_subst[simplified],simp,simp)
  apply(erule wf_equiv_trans)
  apply(rule wf_equiv_eta)
  apply(erule substitution_general)
  apply(assumption)
  apply(erule wf_equiv_wf(2))
  apply(erule valid_sub_valid[simplified])
  apply(simp add:fresh_prod fresh_list_append subst_fresh2)
    (* case 7 *)
  apply(blast intro:wf_equiv_sym)
    (* case 8 *)
  apply(clarsimp)
  apply(subgoal_tac "\<Sigma>,\<Gamma>' \<turnstile> ta \<approx> ta : \<tau>'")
  prefer 2 apply(rule wf_equiv_trans, assumption, erule wf_equiv_sym)
  apply(blast intro:wf_equiv_trans)
  done

lemma wf_equiv_equiv_subst1:
  "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma> \<turnstile> M \<approx> N : \<tau>; \<Sigma>,\<Gamma> \<turnstile> t \<approx> u : \<tau>; valid_ctx \<Gamma>; x \<sharp> \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M{x::=t} \<approx> N{x::=u} : \<tau> "
  apply(erule wf_equiv_equiv_subst[where \<Gamma>'="[]", simplified], auto)
  done


lemma whr_subject_reduction:
  "M \<leadsto> N \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<approx> N : \<tau>"
  apply(nominal_induct M N avoiding: \<Gamma> \<tau> rule:whr.strong_inducts)
  apply(erule wf.cases,simp_all add:hoas.inject)
  apply(erule_tac x="x" in wf.strong_cases,simp_all add:hoas.inject )
  apply(clarify, simp add:abs_fresh hoas.inject hoas_ty.inject alpha,clarify)
  apply(rule wf_equiv_beta)
  apply(blast)
  apply(simp add:fresh_prod)
  apply(erule wf.cases,simp_all add:hoas.inject)
  apply(auto intro: wf_APP wf_equiv_refl)
  done

lemma whr_subject_reduction_ty:
  "M \<leadsto> N \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> N : \<tau>"
  apply(auto intro:whr_subject_reduction wf_equiv_wf)
  done



end


